//////////////////////////////////////////////////////////////////////////
// WB.C -- Whirlybird main source module dealie                         //
//         Whirlybird is a wicked C64 game that I decided to clone in   //
//         an hour.  Then I decided I'd add a bunch of features to make //
//         it a game nearly worth playing.  Thanks to Laura's MIDI      //
//         Heaven for Wagner's Ride of the Valkyries MID, Roland for    //
//         the helicopter and explosion sounds that they put on my      //
//         MIDI sound card, Shawn Hargreaves for Allegro, #tpu'ers for  //
//         not saying it sucks, etc.                                    //
//                                                                      //
//      Things to do:                                                   //
//      High score table                                                //
//      Ending sequence                                                 //
//      ... that's it?  That can't be it.  I must have to do more work  //
//          than that!... hmm... oh, I know!                            //
//      _Add Comments!_  >=]                                            //
//////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <allegro.h>
#include "heli.h"

#define STATE_LIVING 0
#define STATE_EXPLODING 1

typedef struct CHOPPER
{
char curr_frame;
int x, y;
char state;
char behind; // behind background
struct CHOPPER *next;
int nextx, nexty;
int next_frame;
int voice;
} CHOPPER;

CHOPPER *first_chopper;
CHOPPER *last_chopper;

DATAFILE *data;
DATAFILE *expdata;

BITMAP *dblbuf;

int score;
int lives;
int nukes;
int nextchopper;

volatile int ready_for_next_frame;

void allow_frame(void)
{
ready_for_next_frame = 1;
}
END_OF_FUNCTION (allow_frame);

void die(void)
{
BITMAP *expbmp;
BITMAP *bckbmp;
BITMAP *dblbmp;
int index;
int v;

lives--;

expbmp = create_bitmap(80,80); // create bitmap to hold explosion animations
bckbmp = create_bitmap(640,480); // create bitmap to hold background
dblbmp = create_bitmap(640,480); // create double buffer

blit(dblbuf,bckbmp,0,0,0,0,640,480);
v = allocate_voice(data[KABOOM].dat);
voice_start(v);

for(index = 0; index < 64; index += 2)
{
    clear(expbmp);
    draw_rle_sprite(expbmp,expdata[index].dat,0,0);
    blit(bckbmp, dblbmp,0,0,0,0,640,480);
    stretch_sprite(dblbmp, expbmp,80,0,480,480);
    vsync();
    show_mouse(dblbmp);
    blit(dblbmp,screen,0,0,0,0,640,480);
    show_mouse(NULL);
    while(ready_for_next_frame == 0);
    ready_for_next_frame = 0;
}
release_voice(v);
destroy_bitmap(dblbmp);
destroy_bitmap(expbmp);
destroy_bitmap(bckbmp);
}

int find_y(int x)
{
int index;
int finalx;
BITMAP *tmp;

tmp = create_bitmap(1,380);

clear(tmp);
draw_rle_sprite(tmp,data[BACKGROUND].dat,0 - x,-100);

for(index = 0; index<380; index++)
    if(getpixel(tmp,0,index) != 0)
        break;

index += 100;
destroy_bitmap(tmp);
return (index);
}

void init_stuff(void)
{
first_chopper = NULL;
last_chopper = NULL;
dblbuf = create_bitmap(640,480);
data = load_datafile("heli.dat");
expdata = load_datafile("explode.dat");
set_palette(data[GAME_PAL].dat);
set_mouse_sprite(data[CROSSHAIR].dat);
set_mouse_sprite_focus(48,47);
text_mode(-1); // set text to transparent
}

void create_chopper(int x, int y)
{
CHOPPER *tmpchop;

tmpchop = malloc(sizeof(CHOPPER));
tmpchop->behind = 1;
tmpchop->state = STATE_LIVING;
tmpchop->x = x;
tmpchop->y = y;
tmpchop->next = NULL;
tmpchop->curr_frame = 0;
tmpchop->nextx = x;
tmpchop->nexty = find_y(x) - 100;
tmpchop->next_frame = 5;
tmpchop->voice = allocate_voice(data[HELISOUND].dat);
voice_set_playmode(tmpchop->voice,PLAYMODE_LOOP);

if(last_chopper)
{
    last_chopper->next = tmpchop;
    last_chopper = tmpchop;
}
else
    first_chopper = last_chopper = tmpchop;
}

void move_choppers(void)
{
CHOPPER *tmpchop;
int test;

tmpchop = first_chopper;

while(tmpchop)
{
    if(tmpchop->state == STATE_LIVING)
        {
        tmpchop->x += (float)((float)(tmpchop->nextx - tmpchop->x) / (float)tmpchop->next_frame);
        tmpchop->y += (float)((float)(tmpchop->nexty - tmpchop->y) / (float)tmpchop->next_frame);

        if(tmpchop->next_frame == 1)
            {
            tmpchop->behind = 0;
            test = rand()%20;
            if(test == 1)
                {
                tmpchop->nextx = -104;
                tmpchop->nexty = rand()%250 + 100;
                }
            else
                {
            if(test == 2)
                {
                tmpchop->nextx = 640;
                tmpchop->nexty = rand()%250 + 100;
                }              
            else
                {
                tmpchop->nextx = rand()%500 + 10;
                tmpchop->nexty = rand()%250 + 100;
                }
                }
            tmpchop->next_frame = rand()%10 + 10;
            }
        tmpchop->next_frame --;
        }
tmpchop = tmpchop->next;
}
}

void draw_behind_choppers(void)
{
CHOPPER *tmpchop;

tmpchop = first_chopper;
while(tmpchop)
{
if(tmpchop->state == STATE_LIVING && tmpchop->behind == 1)
    {
    draw_sprite(dblbuf, data[HELI1 + tmpchop->curr_frame].dat, tmpchop->x, tmpchop->y);
    if(tmpchop->curr_frame == 0)
        tmpchop->curr_frame ++;
    else
        tmpchop->curr_frame --;
    voice_set_volume(tmpchop->voice,(5-tmpchop->next_frame)*51);
    voice_set_pan(tmpchop->voice,((float)tmpchop->x/(float)2.5));
    voice_start(tmpchop->voice);
    }
tmpchop = tmpchop->next;
}
}

void draw_background(void)
{
draw_rle_sprite(dblbuf,data[BACKGROUND].dat,0,0);
}

void draw_choppers(void)
{
CHOPPER *tmpchop;

tmpchop = first_chopper;
while(tmpchop)
{
if(tmpchop->state == STATE_LIVING && tmpchop->behind == 0)
    {
    draw_sprite(dblbuf, data[HELI1 + tmpchop->curr_frame].dat, tmpchop->x, tmpchop->y);
    if(tmpchop->curr_frame == 0)
        tmpchop->curr_frame ++;
    else
        tmpchop->curr_frame --;
    voice_set_volume(tmpchop->voice,255);
    voice_set_pan(tmpchop->voice,((float)tmpchop->x/(float)2.5));
    voice_start(tmpchop->voice);
    }
else
    if(tmpchop->state != STATE_LIVING)
    {
    draw_rle_sprite(dblbuf, expdata[tmpchop->state - 1].dat, tmpchop->x +10, tmpchop->y - 3);
    tmpchop->state += 2;
    if(tmpchop->state >= 65) // fully exploded
        {
        tmpchop->behind = 1;
        tmpchop->state = STATE_LIVING;
        tmpchop->x = (rand()%400) + 120;
        tmpchop->y = (rand()%200) + 200;
        tmpchop->nextx = tmpchop->x;
        tmpchop->nexty = find_y(tmpchop->x) - 100;
        tmpchop->next_frame = 5;
        deallocate_voice(tmpchop->voice);
        tmpchop->voice = allocate_voice(data[HELISOUND].dat);
        voice_set_playmode(tmpchop->voice,PLAYMODE_LOOP);
        }
    }

tmpchop = tmpchop->next;
}
}

void check_collisions(void)
{
CHOPPER *tmpchop;

tmpchop = first_chopper;

while(tmpchop)
{
if(tmpchop->state == STATE_LIVING && tmpchop->behind == 0)
    if(getpixel(data[HELI1 + tmpchop->curr_frame].dat,mouse_x - tmpchop->x, mouse_y - tmpchop->y) == 0)
        {
        tmpchop->state = STATE_EXPLODING;
        score += 50;
        deallocate_voice(tmpchop->voice);
        tmpchop->voice = allocate_voice(data[KABOOM].dat);
        voice_set_playmode(tmpchop->voice,PLAYMODE_PLAY);
        voice_set_pan(tmpchop->voice,((float)tmpchop->x/(float)2.5));
        voice_start(tmpchop->voice);
        }
if(tmpchop->state == STATE_LIVING)
    if(tmpchop->x < -100 || tmpchop->x > 635)
    {
        die();
        tmpchop->behind = 1;
        tmpchop->state = STATE_LIVING;
        tmpchop->x = (rand()%400) + 120;
        tmpchop->y = (rand()%200) + 200;
        tmpchop->nextx = tmpchop->x;
        tmpchop->nexty = find_y(tmpchop->x) - 100;
        tmpchop->next_frame = 5;
        deallocate_voice(tmpchop->voice);
        tmpchop->voice = allocate_voice(data[HELISOUND].dat);
        voice_set_playmode(tmpchop->voice,PLAYMODE_LOOP);
   }

tmpchop = tmpchop->next;
}
}

void show_stats(void)
{
char stuff[10];
int index;

blit(data[BAR].dat,dblbuf,0,0,20,0,600,50);
sprintf(stuff, "%d", score);
textout(dblbuf,font,stuff,526 - text_length(font,stuff),20,0);
sprintf(stuff, "%d", lives);
textout(dblbuf,font,stuff,590,20,0);
if(score > nextchopper)
{
    nextchopper *= 2;
    create_chopper((rand()%400) + 120,(rand()%200) + 200);
}
for(index = 0; index < nukes; index++)
    draw_rle_sprite(dblbuf,data[NUKE].dat,600 - (index * 30),420);
}

void nuke(void)
{
int index;
int v;
BITMAP *buf;
CHOPPER *tmpchop;

if(nukes == 0) return;
nukes --;

buf = create_bitmap(640,480);

v = allocate_voice(data[KABOOM].dat);
voice_start(v);

for(index = 0; index < 240; index += 6)
{         
    clear(buf);
    rectfill(buf,0,240 - index,639,240 + index,5);
    show_mouse(dblbuf);
    draw_sprite(buf,dblbuf,0,0);
    show_mouse(NULL);
    blit(buf,screen,0,0,0,0,640,480);
}

tmpchop = first_chopper;

while(tmpchop)
{
    tmpchop->behind = 1;
    tmpchop->state = STATE_LIVING;
    tmpchop->x = (rand()%400) + 120;
    tmpchop->y = (rand()%200) + 200;
    tmpchop->nextx = tmpchop->x;
    tmpchop->nexty = find_y(tmpchop->x) - 100;
    tmpchop->next_frame = 5;
    deallocate_voice(tmpchop->voice);
    tmpchop->voice = allocate_voice(data[HELISOUND].dat);
    voice_set_playmode(tmpchop->voice,PLAYMODE_LOOP);


    tmpchop = tmpchop->next;
}
destroy_bitmap(buf);
release_voice(v);
}

void destroy_choppers(void)
{
CHOPPER *tmpchop;
CHOPPER *nxt;

tmpchop = first_chopper;
while(tmpchop)
{
    nxt = tmpchop->next;
    free (tmpchop);
    tmpchop = nxt;
}
}

int main(void)
{
allegro_init();
install_timer();
install_mouse();
install_keyboard();

install_sound(DIGI_AUTODETECT,MIDI_AUTODETECT,NULL);

set_gfx_mode(GFX_AUTODETECT,640,480,0,0);

LOCK_FUNCTION(allow_frame);
LOCK_VARIABLE(ready_for_next_frame);

ready_for_next_frame = 0;

init_stuff();

// Title Screen
clear(screen);
set_palette(black_palette);
draw_rle_sprite(screen,data[TITLE].dat,0,0);
fade_in(data[GAME_PAL].dat,1);
while(!mouse_b && !key[KEY_SPACE] && !key[KEY_ENTER] && !key[KEY_ESC]);
fade_out(1);
clear(screen);

set_palette(data[GAME_PAL].dat);
create_chopper((rand()%400) + 120,(rand()%200) + 200);
create_chopper((rand()%400) + 120,(rand()%200) + 200);

lives = 3;
score = 0;
nukes = 5;
nextchopper = 1000;

play_looped_midi(data[MUSIC].dat,0,289);

install_int(allow_frame, 75);

while (!key[KEY_ESC] && lives > 0)
{
    if(mouse_b)
        nuke();
    move_choppers();
    check_collisions();
    clear(dblbuf);
    draw_behind_choppers();
    draw_background();
    draw_choppers();
    show_stats();

    while(ready_for_next_frame == 0);
    ready_for_next_frame = 0;

    show_mouse(dblbuf);
    vsync();
    blit(dblbuf,screen,0,0,0,0,640,480);
    show_mouse(NULL);
}
remove_int(allow_frame);
for(nukes=0;nukes < 64; nukes++)
    voice_ramp_volume(nukes, 500, 0);
midi_loop_end = -1;
midi_loop_start = -1;
midi_seek(290);
while(midi_pos > 280);
stop_midi();
fade_out(1);
destroy_bitmap(dblbuf);
destroy_choppers();
unload_datafile(data);
unload_datafile(expdata);
set_gfx_mode(GFX_TEXT,80,25,0,0);
printf("Final score: %d\n\n",score);
return(0);
}
